﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.Knapsack;
using HeuristicLab.Problems.TravelingSalesman;

namespace HeuristicLab.Algorithms.ScatterSearch.TravelingSalesman {
  /// <summary>
  /// An operator that improves traveling salesman solutions.
  /// </summary>
  [Item("TravelingSalesmanImprovementOperator", "An operator that improves traveling salesman solutions.")]
  [StorableClass]
  public sealed class TravelingSalesmanImprovementOperator : SingleSuccessorOperator, ILocalImprovementOperator, IScatterSearchTargetProcessor {
    #region Problem properties
    public Type ProblemType {
      get { return typeof(KnapsackProblem); }
    }
    [Storable]
    private KnapsackProblem problem;
    public IProblem Problem {
      get { return problem; }
      set { problem = (KnapsackProblem)value; }
    }
    #endregion

    #region Parameter properties
    public ScopeParameter CurrentScopeParameter {
      get { return (ScopeParameter)Parameters["CurrentScope"]; }
    }
    public IValueLookupParameter<DoubleMatrix> CoordinatesParameter {
      get { return (IValueLookupParameter<DoubleMatrix>)Parameters["Coordinates"]; }
    }
    public IValueLookupParameter<IEvaluator> EvaluatorParameter {
      get { return (IValueLookupParameter<IEvaluator>)Parameters["Evaluator"]; }
    }
    public IValueLookupParameter<IntValue> ImprovementAttemptsParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["ImprovementAttempts"]; }
    }
    public IValueLookupParameter<IRandom> RandomParameter {
      get { return (IValueLookupParameter<IRandom>)Parameters["Random"]; }
    }
    public IValueLookupParameter<IItem> TargetParameter {
      get { return (IValueLookupParameter<IItem>)Parameters["Target"]; }
    }
    #region ILocalImprovementOperator Parameters
    public IValueLookupParameter<IntValue> MaximumIterationsParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["MaximumIterations"]; }
    }
    public ILookupParameter<IntValue> EvaluatedSolutionsParameter {
      get { return (ILookupParameter<IntValue>)Parameters["EvaluatedSolutions"]; }
    }
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    #endregion
    #endregion

    #region Properties
    public IScope CurrentScope {
      get { return CurrentScopeParameter.ActualValue; }
    }
    public DoubleMatrix Coordinates {
      get { return CoordinatesParameter.ActualValue; }
      set { CoordinatesParameter.ActualValue = value; }
    }
    public IEvaluator Evaluator {
      get { return EvaluatorParameter.ActualValue; }
      set { EvaluatorParameter.ActualValue = value; }
    }
    public IntValue ImprovementAttempts {
      get { return ImprovementAttemptsParameter.ActualValue; }
      set { ImprovementAttemptsParameter.ActualValue = value; }
    }
    public IRandom Random {
      get { return RandomParameter.ActualValue; }
      set { RandomParameter.ActualValue = value; }
    }
    private IItem Target {
      get { return TargetParameter.ActualValue; }
    }
    #endregion

    [StorableConstructor]
    private TravelingSalesmanImprovementOperator(bool deserializing) : base(deserializing) { }
    private TravelingSalesmanImprovementOperator(TravelingSalesmanImprovementOperator original, Cloner cloner)
      : base(original, cloner) {
      this.problem = cloner.Clone(original.problem);
    }
    public TravelingSalesmanImprovementOperator()
      : base() {
      #region Create parameters
      Parameters.Add(new ScopeParameter("CurrentScope"));
      Parameters.Add(new ValueLookupParameter<DoubleMatrix>("Coordinates"));
      Parameters.Add(new ValueLookupParameter<IEvaluator>("Evaluator"));
      Parameters.Add(new ValueLookupParameter<IntValue>("ImprovementAttempts", new IntValue(100)));
      Parameters.Add(new ValueLookupParameter<IRandom>("Random"));
      Parameters.Add(new ValueLookupParameter<IItem>("Target"));
      #endregion
      TargetParameter.ActualName = "TSPTour"; // temporary solution for the traveling salesman problem
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new TravelingSalesmanImprovementOperator(this, cloner);
    }

    public override IOperation Apply() {
      var currSol = CurrentScope.Variables[TargetParameter.ActualName].Value as Permutation;
      var bestSol = currSol;
      double currLength = TSPEuclideanPathEvaluator.Apply(Evaluator as TSPCoordinatesPathEvaluator, Coordinates, currSol);
      double bestLength = currLength;

      for (int i = 0; i < ImprovementAttempts.Value; i++) {
        int a = Random.Next(currSol.Length - 1);
        int b = Random.Next(currSol.Length - 1);
        Invert(currSol, a, b);
        currLength = TSPEuclideanPathEvaluator.Apply(Evaluator as TSPCoordinatesPathEvaluator, Coordinates, currSol);
        if (currLength < bestLength) {
          bestLength = currLength;
          bestSol = (Permutation)currSol.Clone();
        }
        Invert(currSol, a, b);
      }

      CurrentScope.Variables[TargetParameter.ActualName].Value = bestSol;

      return base.Apply();
    }

    private void Invert(Permutation sol, int i, int j) {
      if (i != j) {
        for (int a = 0; a < Math.Abs(i - j) / 2; a++) {
          int tmp = sol[(i + a) % sol.Length];
          sol[(i + a) % sol.Length] = sol[(j - a + sol.Length) % sol.Length];
          sol[(j - a + sol.Length) % sol.Length] = tmp;
        }
      }
    }
  }
}
